@@ -9,9 +9,9 @@ import warnings
 from os.path import splitext
 
 import numpy as np
+from numpy.lib.arraysetops import unique
 from scipy.constants import speed_of_light
 from scipy.optimize import OptimizeWarning, least_squares
-from scipy.optimize._numdiff import approx_derivative
 from scipy.stats import norm
 from tqdm import tqdm
 
@@ -22,10 +22,11 @@ from .atmosphere.krzfile import KrzFile
 from .atmosphere.savfile import SavFile
 from .large_file_storage import setup_lfs
 from .nlte import DirectAccessFile
-from .sme_synth import SME_DLL
 from .synthesize import Synthesizer
 from .util import print_to_log
 
+from ._numdiff import approx_derivative
+
 logger = logging.getLogger(__name__)
 
 clight = speed_of_light * 1e-3  # km/s
@@ -34,13 +35,9 @@ warnings.filterwarnings("ignore", category=OptimizeWarning)
 
 class SME_Solver:
     def __init__(self, filename=None, restore=False):
-        self.dll = SME_DLL()
         self.config, self.lfs_atmo, self.lfs_nlte = setup_lfs()
         self.synthesizer = Synthesizer(
-            config=self.config,
-            lfs_atmo=self.lfs_atmo,
-            lfs_nlte=self.lfs_nlte,
-            dll=self.dll,
+            config=self.config, lfs_atmo=self.lfs_atmo, lfs_nlte=self.lfs_nlte,
         )
 
         # Various parameters to keep track of during solving
@@ -52,8 +49,8 @@ class SME_Solver:
         self.restore = restore
 
         # For displaying the progressbars
-        self.progressbar = None
-        self.progressbar_jacobian = None
+        # self.progressbar = None
+        # self.progressbar_jacobian = None
 
     @property
     def nparam(self):
@@ -130,6 +127,8 @@ class SME_Solver:
         save = not isJacobian and self.filename is not None
         reuse_wavelength_grid = isJacobian
         radial_velocity_mode = "robust" if not isJacobian else "fast"
+        # method = "parallel" if isJacobian else "sequential"
+        method = "sequential"
 
         # change parameters
         for name, value in zip(self.parameter_names, param):
@@ -144,6 +143,7 @@ class SME_Solver:
                 passLineList=False,
                 updateLineList=self.update_linelist,
                 radial_velocity_mode=radial_velocity_mode,
+                method=method,
             )
         except AtmosphereError as ae:
             # Something went wrong (left the grid? Don't go there)
@@ -171,11 +171,11 @@ class SME_Solver:
         resid = np.nan_to_num(resid, copy=False)
 
         # Update progress bars
-        if isJacobian:
-            self.progressbar_jacobian.update(1)
-        else:
-            self.progressbar.total += 1
-            self.progressbar.update(1)
+        # if isJacobian:
+        #     self.progressbar_jacobian.update(1)
+        # else:
+        #     self.progressbar.total += 1
+        #     self.progressbar.update(1)
 
         if not isJacobian:
             # Save result for jacobian
@@ -196,11 +196,14 @@ class SME_Solver:
         The calculation is the same as "3-point"
         but we can tell residuals that we are within a jacobian
         """
-        self.progressbar_jacobian.reset()
+        # self.progressbar_jacobian.reset()
+
+        # Here we replace the scipy version of approx_derivative with our own
+        # The only difference being that we use Multiprocessing for the jacobian
         g = approx_derivative(
             self.__residuals,
             param,
-            method="3-point",
+            method="2-point",
             # This feels pretty bad, passing the latest synthetic spectrum
             # by reference as a parameter of the residuals function object
             f0=self._latest_residual,
@@ -257,12 +260,12 @@ class SME_Solver:
             or "monh" in self.parameter_names
         ):
             if sme.atmo.method == "grid":
-                atmo_file = sme.atmo.source
-                _, ext = splitext(atmo_file)
-                atmo_file = self.lfs_atmo.get(atmo_file)
+                atmo_source = sme.atmo.source
+                _, ext = splitext(atmo_source)
+                atmo_file = self.lfs_atmo.get(atmo_source)
 
                 if ext == ".sav":
-                    atmo_grid = SavFile(atmo_file)
+                    atmo_grid = SavFile(atmo_file, source=atmo_source)
 
                     teff = np.unique(atmo_grid.teff)
                     teff = np.min(teff), np.max(teff)
@@ -278,7 +281,7 @@ class SME_Solver:
                 elif ext == ".krz":
                     # krz atmospheres are fixed to one parameter set
                     # allow just "small" area around that
-                    atmo = KrzFile(atmo_file)
+                    atmo = KrzFile(atmo_file, source=atmo_source)
                     bounds["teff"] = atmo.teff - 500, atmo.teff + 500
                     bounds["logg"] = atmo.logg - 1, atmo.logg + 1
                     bounds["monh"] = atmo.monh - 1, atmo.monh + 1
@@ -630,7 +633,7 @@ class SME_Solver:
 
         # Divide the uncertainties by the spectrum, to improve the fit in the continuum
         # Just as in IDL SME, this increases the relative error for points inside lines
-        # uncs /= spec
+        uncs /= spec
 
         logger.info("Fitting Spectrum with Parameters: %s", ",".join(param_names))
         logger.debug("Initial values: %s", p0)
@@ -650,13 +653,14 @@ class SME_Solver:
             )
 
         # Setup LineList only once
-        self.dll.SetLibraryPath()
-        self.dll.InputLineList(sme.linelist)
+        dll = self.synthesizer.get_dll()
+        dll.SetLibraryPath()
+        dll.InputLineList(sme.linelist)
 
         # Do the heavy lifting
         if self.nparam > 0:
-            self.progressbar = tqdm(desc="Iteration", total=0)
-            self.progressbar_jacobian = tqdm(desc="Jacobian", total=len(p0) * 2)
+            # self.progressbar = tqdm(desc="Iteration", total=0)
+            # self.progressbar_jacobian = tqdm(desc="Jacobian", total=len(p0))
             with print_to_log():
                 res = least_squares(
                     self.__residuals,
@@ -664,15 +668,15 @@ class SME_Solver:
                     jac=self.__jacobian,
                     bounds=bounds,
                     x_scale="jac",
-                    loss="linear",
+                    loss="soft_l1",
                     method="trf",
                     verbose=2,
                     max_nfev=sme.fitresults.maxiter,
                     args=(sme, spec, uncs, mask),
                     kwargs={"bounds": bounds, "segments": segments},
                 )
-            self.progressbar.close()
-            self.progressbar_jacobian.close()
+            # self.progressbar.close()
+            # self.progressbar_jacobian.close()
             # The returned jacobian is "scaled for robust loss function"
             res.jac = self._last_jac
             for i, name in enumerate(self.parameter_names):
